package master;

import org.locationtech.jts.geom.Coordinate;
import org.locationtech.jts.geom.GeometryFactory;
import org.locationtech.jts.geom.Point;
import org.locationtech.jts.geom.Polygon;
import org.locationtech.jts.geom.prep.PreparedGeometry;
import org.locationtech.jts.geom.prep.PreparedPolygon;

import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentNavigableMap;


public class Filter implements Callable<RefinedFeatures> {

    ConcurrentNavigableMap<long[], String> features;
    PreparedGeometry poly;
    LocalDateTime date_min;
    LocalDateTime date_max;
    boolean temporal;


    public Filter(ConcurrentNavigableMap<long[], String> features) {

        this.features = features;
        this.temporal = false;

    }

    public void setTemporalPredicate(LocalDateTime date_min, LocalDateTime date_max) {

        this.temporal = true;
        this.date_min = date_min;
        this.date_max = date_max;

    }

    public void setSpatialPredicate(Polygon poly) {
        this.poly = new PreparedPolygon(poly);
    }

    @Override
    public RefinedFeatures call() {

        Collection<String> valid_features = new ArrayList<>();
        GeometryFactory geometryFactory = new GeometryFactory();
        DateTimeFormatter formatter = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss");
        long false_positive_count = 0;
        long true_positive_count = 0;


        for (long[] key : features.keySet()) {
            String feat = features.get(key);
            String[] feature_string = feat.split(",");
            String lat = feature_string[0];
            String lon = feature_string[1];
            Coordinate coord = new Coordinate(Double.parseDouble(lat), Double.parseDouble(lon));
            Point point = geometryFactory.createPoint(coord);

            if (this.temporal) {
                LocalDateTime date = LocalDateTime.parse(feature_string[5] + " " + feature_string[6], formatter);

                if (date.isAfter(date_min) && date.isBefore(date_max) && poly.intersects(point)) {
                    valid_features.add(feat);
                    true_positive_count++;
                } else {
                    false_positive_count++;
                }
            } else {
                if (poly.intersects(point)) {
                    valid_features.add(feat);
                    true_positive_count++;
                } else {
                    false_positive_count++;
                }
            }

        }
        return new RefinedFeatures(valid_features, false_positive_count, true_positive_count);

    }

}
