package master;

import com.google.common.base.Strings;
import it.unimi.dsi.fastutil.longs.Long2DoubleMap;
import it.unimi.dsi.fastutil.longs.Long2DoubleOpenHashMap;
import org.mapdb.BTreeMap;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

public class FrequencyPrefixHistogram {


    Long2DoubleOpenHashMap frequency_prefix_histogram_map;
    Long2DoubleOpenHashMap epoch_frequency_histogram_map;
    int sample_number;
    long bit_precision;
    int dimensions;
    int bits_per_dimension;
    boolean normalize;


    public FrequencyPrefixHistogram(int sample_number, long bit_precision, boolean normalize) {


        this.sample_number = sample_number;
        this.epoch_frequency_histogram_map = new Long2DoubleOpenHashMap();
        this.frequency_prefix_histogram_map = new Long2DoubleOpenHashMap();
        this.bit_precision = bit_precision;

    }

    public void createFrequencyTable(BTreeMap<long[], String> map, int dimensions, int bitsPerDimension, boolean use_epoch) {


        List<long[]> zkeys = new ArrayList<>();
        zkeys.addAll(map.getKeys());
        Collections.shuffle(zkeys);

        this.dimensions = dimensions;
        this.bits_per_dimension = bitsPerDimension;

        Long2DoubleOpenHashMap frequency_map = new Long2DoubleOpenHashMap();
        Long2DoubleOpenHashMap epoch_frequency_map = new Long2DoubleOpenHashMap();


        this.bit_precision = this.bit_precision * dimensions;

        for (long[] key : zkeys.subList(0,  sample_number)) {

            String bin_zkey_epoch = Strings.padStart(Long.toBinaryString((long) key[1]), dimensions * bitsPerDimension, '0');
            String bin_zkey = Long.toBinaryString( key[0]);


            for (int i = dimensions; i < dimensions * bitsPerDimension - this.bit_precision; i += dimensions) {
                if (dimensions == 3 && use_epoch) {

                    String bin_sub_zkey_string = bin_zkey_epoch.substring(0, i);
                    String bin_sub_zkey_trimmed = Strings.padEnd(bin_sub_zkey_string, this.dimensions * this.bits_per_dimension - (int) this.bit_precision, '0');
                    String bin_epoch = Strings.padStart(Long.toBinaryString((long) key[0]), (int) this.bit_precision, '0');

                    long bin_sub_epoch_zkey_test = Long.parseLong(bin_epoch + bin_sub_zkey_trimmed, 2);
                    double count = epoch_frequency_map.getOrDefault(bin_sub_epoch_zkey_test, 0);
                    epoch_frequency_map.put(bin_sub_epoch_zkey_test, count + 1);

                } else {
                    String bin_sub_zkey = bin_zkey.substring(0, i);
                    long bin_sub_zkey_long = Long.parseLong(Strings.padEnd(bin_sub_zkey, bin_zkey.length(), '0'), 2);
                    double count = frequency_map.getOrDefault(bin_sub_zkey_long, 0);
                    frequency_map.put(bin_sub_zkey_long, count + 1);

                }
            }

        }

        if(normalize) {
            if (dimensions == 3 && use_epoch) {
                normalizeEpochFrequencyTable(epoch_frequency_map);
            } else {
                normalizeFrequencyTable(frequency_map);
            }
        }
        else {
            this.epoch_frequency_histogram_map = epoch_frequency_map;
            this.frequency_prefix_histogram_map = frequency_map;
        }

        this.frequency_prefix_histogram_map.trim();
        this.epoch_frequency_histogram_map.trim();




    }

    public void normalizeEpochFrequencyTable(Long2DoubleOpenHashMap dict) {


        Long2DoubleMap factor_map = new Long2DoubleOpenHashMap();
        for (long key : dict.keySet()) {

            String binary_key = Strings.padStart(Long.toBinaryString(key), this.dimensions * this.bits_per_dimension, '0');
            String epoch_binary = binary_key.substring(0, (int) this.bit_precision - 1);
            long epoch = (Long.parseLong(epoch_binary, 2));

            double count = factor_map.getOrDefault(epoch, 0);
            factor_map.put(epoch, count + 1);


        }

        for (long key : dict.keySet()) {


            String binary_key = Strings.padStart(Long.toBinaryString(key), this.dimensions * this.bits_per_dimension, '0');
            String epoch_binary = binary_key.substring(0, (int) this.bit_precision - 1);
            long epoch = (Long.parseLong(epoch_binary, 2));
            double normmalized_count = dict.get(key) / factor_map.get(epoch);
            this.epoch_frequency_histogram_map.put(key, normmalized_count);

        }

    }

    public void normalizeFrequencyTable(Long2DoubleOpenHashMap dict) {
        long sum = 0;
        for (double v : dict.values()) {
            sum += v;
        }
        double factor = 1.0 / (double) sum;
        for (long key : dict.keySet()) {
            double count = dict.get(key);
            double norm_count = factor * (double) count;
            this.frequency_prefix_histogram_map.put(key, norm_count);
        }
    }


}
