package master;

import org.mapdb.BTreeMap;
import org.mapdb.DB;
import org.mapdb.DBMaker;
import org.mapdb.Serializer;

import java.io.IOException;
import java.util.List;


public class Main {


    public static void main(String[] args) throws IOException, InterruptedException {




        char SET_PERIOD;
        String SET_GEOLIFE_DATA_LOCATION;
        int SET_BITS_PER_DIMENSIONS;
        boolean SET_LOAD_DATA;
        int SET_FREQUENCY_PREFIX_HISTOGRAM_SAMPLE_NUMBER;
        int SET_FREQUENCY_PREFIX_HISTOGRAM_PRECISION;
        boolean SET_NORMALIZED_FREQUENCY_PREFIX_HISTOGRAM;
        boolean SET_MULTITHREADING;
        int SET_CPU_CORES;
        boolean SET_EPOCH;
        boolean SET_TEMPORAL;
        boolean SET_BEST_FIRST_TRAVERSAL;
        int SET_DIMENSIONS;
        boolean SET_FETCH_DATA;
        int SET_MAX_RANGES;

        //////////////////////////////
        //////// PARAMETERS //////////
        SET_EPOCH = true;
        SET_TEMPORAL = true;
        SET_BEST_FIRST_TRAVERSAL = true;
        SET_PERIOD = 'y';
        SET_GEOLIFE_DATA_LOCATION = "E:\\Geodata\\geolife";
        SET_LOAD_DATA = false;
        SET_FREQUENCY_PREFIX_HISTOGRAM_SAMPLE_NUMBER = 1000000;
        SET_FREQUENCY_PREFIX_HISTOGRAM_PRECISION = 3; // must  >2 for semi-integrated index.. higher number -> lower precision
        SET_NORMALIZED_FREQUENCY_PREFIX_HISTOGRAM = false;
        SET_MULTITHREADING = true;
        SET_CPU_CORES = 4;
        SET_FETCH_DATA = true;
        SET_MAX_RANGES = 100000;
        //////////////////////////////
        //////////////////////////////


        SET_DIMENSIONS = 2;
        SET_BITS_PER_DIMENSIONS = 31;

        if (SET_TEMPORAL) {
            SET_DIMENSIONS = 3;
            SET_BITS_PER_DIMENSIONS = 21;
        }



        Loader loader = new Loader(SET_GEOLIFE_DATA_LOCATION, SET_DIMENSIONS, SET_BITS_PER_DIMENSIONS);
        List<Query> queries = loader.loadQueries(SET_TEMPORAL);

        List<Query> queries_spatial = loader.loadQueries(false);
        List<Query> queries_spatialtemporal = loader.loadQueries(true);
        //Benchmark benchmark = new Benchmark(queries_spatial, queries_spatialtemporal);
        //benchmark.runBenchmark();


        String file;
        if (SET_TEMPORAL && SET_EPOCH) {
            if (SET_PERIOD == 'w') {
                file = "fileweek.db";
            } else if (SET_PERIOD == 'm') {
                file = "filemonth.db";
            } else {
                file = "fileyear_test2_2.db";
            }
            loader.enableEpoch(SET_PERIOD);
        } else if (SET_TEMPORAL && !SET_EPOCH) {
            file = "file_temporal_test.db";
        } else {
            file = "file_test.db";
        }


        DB db = DBMaker
                .fileDB(file)
                .fileMmapEnable()
                .fileMmapEnableIfSupported()
                .fileMmapPreclearDisable()
                .cleanerHackEnable()
                .make();


        BTreeMap<long[], String> map;

        if (SET_TEMPORAL && SET_EPOCH) {
            map = db.treeMap("map")
                    .keySerializer(Serializer.LONG_ARRAY)
                    .valueSerializer(Serializer.STRING)
                    .valuesOutsideNodesEnable()
                    .createOrOpen();
        } else {
            map = db.treeMap("map")
                    .keySerializer(Serializer.LONG_ARRAY)
                    .valueSerializer(Serializer.STRING)
                    .valuesOutsideNodesEnable()
                    .createOrOpen();
        }



        if (SET_LOAD_DATA) {
            loader.loadData(map);
            //map.putAll(data);
            System.out.println("Start commmiting data");
            db.commit();
            System.out.println("Data succesfully commited\n");
        }




        QueryExecutioner queryExecutioner = new QueryExecutioner(SET_DIMENSIONS, SET_BITS_PER_DIMENSIONS, SET_BEST_FIRST_TRAVERSAL, map);

        if (SET_EPOCH) {
            queryExecutioner.enableEpoch(SET_PERIOD);
        }


        if (SET_BEST_FIRST_TRAVERSAL) {
            FrequencyPrefixHistogram FrequencyPrefixHistogram = new FrequencyPrefixHistogram(SET_FREQUENCY_PREFIX_HISTOGRAM_SAMPLE_NUMBER, SET_FREQUENCY_PREFIX_HISTOGRAM_PRECISION, SET_NORMALIZED_FREQUENCY_PREFIX_HISTOGRAM);
            FrequencyPrefixHistogram.createFrequencyTable(map, SET_DIMENSIONS, SET_BITS_PER_DIMENSIONS, SET_EPOCH);
            queryExecutioner.setFrequencyPrefixHistogram(FrequencyPrefixHistogram);
            System.out.println("Histogram succesfully created\n");

        }

        if (SET_MULTITHREADING) {
            queryExecutioner.enableMultiThreading(SET_CPU_CORES);
        }


		// TEST //
        long time = 0;
        double nr_queries = 0;
        double false_discovery = 0;
        long total_points = 0;
        long parsing_time = 0;
        long planning_time = 0;
        long nubmer_rangers = 0;
        long execution_time = 0;

        for (Query query : queries) {
            Logger logger = new Logger(0);

            if (SET_TEMPORAL) {
                queryExecutioner.executeSpatiotemporalQuery(query, SET_MAX_RANGES, SET_FETCH_DATA, logger);
            } else {
                queryExecutioner.executeSpatialQuery(query, SET_MAX_RANGES, SET_FETCH_DATA, logger);

            }

            if (logger.count == 0) {
                System.out.println(query.getTemporalPredicate()[0] + "--" + query.getTemporalPredicate()[1]);
            }
            time += logger.execution_time + logger.planning_time;
            false_discovery += logger.falseDiscoveryRate();
            nr_queries++;
            total_points += logger.count;
            parsing_time += logger.parsing_time;
            planning_time += logger.planning_time;
            nubmer_rangers += logger.ranges_generated;
            execution_time += logger.execution_time;


        }

        System.out.println("Number of queries: " + nr_queries);
        System.out.println("Avg. false discovery rate: " + (false_discovery / nr_queries));
        System.out.println("Total points: " + total_points);
        System.out.println("Total parsing time: " + parsing_time);
        System.out.println("Total response time: " + time);
        System.out.println("Execution time: " + execution_time);
        System.out.println("Total decomposition time: " + planning_time);
        System.out.println("Avg. ranges created: " + nubmer_rangers / nr_queries);


        db.close();

    }
}

